open Ast
open Past
open Translate_env
open Typechecker

(*This file translates a NwQL syntax tree to a python syntax tree and in the process of this conversion it also performs semantic analysis*)

(************************************************************************
Name=translate_datatype

Description= This method is used to translate datatype to pdatatype

Input= A datatype from AST

Output= A past equivalent datatype*)

let translate_datatype= function 
	Int -> P_int
	|
	String -> P_string
	|
	Boolean -> P_boolean
	|
	AVP -> P_avp

(************************************************************************
Name=translate_expr

Description= This method is used to translate Expr to P_expr

Input=Symbol table and an expression of one of the six types

Output=A tuple based on the updated environment and the translated
P_expr. Note the envrionment is only updated when an expression of
type assignment is being matched, otherwise the received environment
as an input is returned as the output
*************************************************************************)

let rec translate_expr env expr= match expr with
(*translating expression of type IntLit to P_intLit*)
IntLit(i) -> (P_intLit(i),env)
|
(*translating expression of type StrLit to P_strLit*)
StrLit(s) -> (P_strLit(s),env)
|
(*translating expression of type BoolLit to P_boolLit*)
BoolLit(b) -> (P_boolLit(b),env)
|
(*translating expression of type AVPList to P_avpList*)
AVPList(l) -> 
	(P_avpList(l),env)
|
(*translating expression of type AVPList to P_avpList*)
Id(id) -> (P_id(id),env)
|
(*translating expression of type LitAssign to P_litAssign*)
LitAssign(name,value) as expr->
        (*check if the assignment expression is consistent with its type*)
        (*check assignment will return the updated environment*)
        let updatedenv= (check_assigexpr env expr)
        and
        (*here we call the translate_expr again to translate the literal to p_literal and catch the translated object in a variable*)
        (val_to_pval,_) = (translate_expr env value) in
        (*here we return passignment with the updated environment*)
        ((P_litAssign(name, val_to_pval)),updatedenv)

(************************************************************************
Name=translate_gvdecl                                    

Description= translate global variable declarations 
to python global varible declarations

Input= A symbol tablex-which would be typically empty at
at the time of declaration-and a global variable declaration
to be translated

Output= A tuple based on translated global variable and update symbol 
table 
*************************************************************************)

let translate_gvdecl env gvar=
(*first check if the variable exist*)
        if gvar_exist gvar.vname env then
                (*if the variable exists then raise an exception that the variable is already defined*)
                raise(Failure("Variable " ^ gvar.vname ^" already defined"))
		(*if the variable does not exist check if the value of the variable is consistent with its type*)
        else    
                if check_gvdecl_consistency gvar then
                (*if the value is consistent then update the symbol table and call the translate expression method since the assigned literal is an expression*)
                        let updenv=NameMap.add  gvar.vname gvar env in 
		(*translate expression will return the translated p_expression*)
			let (pexp,_)=translate_expr updenv gvar.value
		(*use the returned p_expression as well as the translate_datatype method to create pvdecl type object*)
			in let pgvdecl_trans={pvtype=(translate_datatype gvar.vtype); pvname=gvar.vname;pvalue=pexp}
		(*return pvdecl type object and the update symbol table*)
			in (pgvdecl_trans,updenv)
                else
                (*if the value is not consistent with its type then fail with the exception that the value is incompatible*)
                         raise(Failure("Incompatible value for " ^ gvar.vname))

(*********************************************************************************************
Name=translate_stmt

Description= Translate statements to P_statements

Input = A type of statement

Output= Equivalent type of p_statement
*************************************************************************)

(*translating statement of type expressions*)
let translate_stmt env stmts = 
	match stmts  with  Expr(expr) -> 
		(*translate_expr returns P_expr and an update environment, out of which we create P_expr*)
		let expr,env=translate_expr env expr in (P_expr(expr),env)
	|
	Select(col_list,tab_name) ->
		if ((check_colnames col_list) = true) && ((check_tablename tab_name) = true)                        
			then (Pselect(col_list,tab_name),env)
                else
                        raise(Failure("Illegal table name " ^ tab_name))
	|
	Insert(tab_name,col_list,exprs) ->
                (*We check if tablename is correct, mandatory columnnames are present, the optional columnnames are correct and no duplicate columnnames are there*)
                if ((check_tablename tab_name)=true)&&((check_insert_colnames col_list)=true)&&((check_colnames col_list)=true)&&((find_dup col_list)=false)
                        then
                                (*now check if number of columnnames is greater than the number of expressions*)
                                if (List.length col_list) > (List.length exprs)
                                        then
                                                (*if number of columnnames is greater than the number of expressions then raise an exception*)
                                                raise((Failure("No value for column " ^ (List.nth col_list ((List.length exprs))))))
                                (*check if the number of columnnames are less than the the number of expressions*)
                                else if (List.length col_list) < (List.length exprs)
                                        then
                                                (*if number of columnnames are less than number of expressions then raise an exception*)
                                                raise ((Failure("The expression " ^ ptyexpr (List.nth exprs ((List.length col_list) ))^ " does not belong to any column")))
 				else
                                        (*If number of columnnames are equal to number of expressions then*)
						(*we combine the list of columnnames and list of expressions in to a list of tuples*)
                                        let col_exprs_combine = List.combine col_list exprs in
					(*In the combined list we add environment to each tuples since check_value takes a tuple of env,column and value*)
                                                let rec env_col_exprs_combine= function [] -> []
                                                |
                                                hd::tail -> (env,fst hd, snd hd) ::  env_col_exprs_combine tail in
                                                        let combined_list= env_col_exprs_combine col_exprs_combine in
								(*Apply list.map to each of the elements of the combined list*)
								(*We can simply ignore the resulting list since if the semantic check are not passed 
								then an exception will be raised by check_value*)
								let _=List.map check_value combined_list in
									(*List.map concluding successfully means that check_value didnt cause exception*)
									(*first we are going to translate a list of Expr to Pexpr*)
									(*since env is the same we create a partial application of a function*)
									let expr_map=translate_expr env in
									(*this partially applied function will take an expression*)
									let p_expr_list = List.map expr_map exprs in
									(*apply the partial function to a list of expression that would retrun a list of
									p_expr and environment*)
									let p_exprs=List.map fst p_expr_list in
									(*since Pinsert only requires pexpressions we apply first to be above returned list*)
									(*creating and returning pinsert*)
									(Pinsert(tab_name,col_list,p_exprs),env)
		(*if tablename is not valid an exception will be raised*)
		else
			raise(Failure("Illegal table name " ^ tab_name))
	|
	(*Translating Delete from ast to P_delete in past*)
	Delete(fnamelist,nodename,tab_name) ->
		(*first we check if table name correct and if there any duplicate expressions*)
		if((check_tablename tab_name)=true) && ((find_dup_expr fnamelist)=false)
			then
				(*if yes then we serialise flow identifier name, columnname and environment so that we can pass it to typechecker*)
				let join elem1 elem2 elem3 = (elem1,elem2,elem3) in
					let par_join expr=join env "name" expr in
						let env_cname_exp_list=List.map par_join fnamelist in
							(*if the exrepssions are not of type Id or StrLit matching the name regexp then exceptions will be raised*)
							let _=List.map check_value env_cname_exp_list and
							_ = check_value (env,"node",nodename) in
							(*If the above List.map is successful then semantic analysis is successful so far*)
							(*now translating the expression to p_expr*)
								let expr_map=translate_expr env in
									let p_expr_list=List.map expr_map fnamelist in
										let p_exprs= List.map fst p_expr_list in
										let p_node= fst (expr_map nodename) in
											(Pdelete(p_exprs,p_node,tab_name),env)
		else
			raise(Failure("Illegal table name " ^ tab_name))
						
			


(************************************************************************
Name=translate_program

Description= Translate a program to p_program

Input= A tuple based on a list of global variable declarations 
and a list of statements

Output= A list of translated p_global variable declarations and 
a list of p_statements
*************************************************************************)

let translate_program (gvdecllist,stmtlist) =
	
	(************************************************************************
	Name=trav_stmts
	
	Description=Translate statements to P_statements

	Input= A program environment and a list of statements

	Output = A list of P_statements 
	*************************************************************************)
	let rec trav_stmts env stmts=
	match stmts with
	hd::tail ->
		(*Translate head of the list and store result*)
		let result = translate_stmt env hd in
		(*Append the p_statement to a list whose tail calls traverse statements with the updated environment receved in the result*)
			(fst result) :: trav_stmts (snd result) tail
	|
	[] -> []

	(************************************************************************
	Name=get_pgvdecl_updtbl
	
	Description=Translate gvdecl to pgvdecl

	Input= An environment and a list of global declarations

	Output= A list of pgvdecl
	*************************************************************************)

	and get_pgvdecl_updtbl env gvdecl = match gvdecl with
	(*Translate head of the list and store result*)
	hd::tail -> 
		let result = translate_gvdecl env hd in
	(*Append the result including both pgvdecl and environment to a list whose tail calls traverse pgvdecl with the updated environment*)
			result :: get_pgvdecl_updtbl (snd result) tail
	|
	[] -> []

		in
		(*Initialise an empty symbol tablr*) 
		let init_symtab = NameMap.empty in
			(*call the  get_pgvdecl_updtbl method*)
			let comb_list = get_pgvdecl_updtbl init_symtab gvdecllist in
			(*separate the pgvdecl into a list, from a list of (pgvdecl,env)*)
				let pgvdecl_list =List.map fst comb_list
			(*from combined list take out the most recent runtime environment*)
			and upd_sym_tbl =  snd (List.hd (List.rev comb_list))
			in
				(*Call the trav_stmts methods with the updated runtime and statement list 
				and append the resulting list as a second member of a tuple
				while first being the pgvdecl list*)
				pgvdecl_list,trav_stmts upd_sym_tbl stmtlist
