open Ast
open Translate_env
module Regex = Re2.Regex


(************************************************************************
Name=check_gvdecl_consistency

Description= Checks if the value of a global variable is consistent
with its type

Input= gvdel record type

Output= true/false
*************************************************************************)

let check_gvdecl_consistency gvdecl= match gvdecl.vtype with
	Int -> let checkvalue value= match value with
		IntLit(_) -> true
		|
		_ -> false
		in checkvalue gvdecl.value
	|
	String -> let checkvalue value= match value with
		StrLit(_) -> true
		|
		_ -> false
		in checkvalue gvdecl.value
	|

	Boolean -> let checkvalue value= match value with
		BoolLit(_) -> true
		|
		_ -> false
		in checkvalue gvdecl.value
	|

	AVP -> let checkvalue value= match value with
		AVPList(_) -> true
	|
		_ -> false
		in checkvalue gvdecl.value

(************************************************************************
Name = check_assigexpr

Description= This method is used for typechecking assignment expressions

Input = An environment and an expression

Output = Updated environment
*************************************************************************)
let check_assigexpr env expr =
	(*Match only the Literal assignment*)
	match expr with LitAssign(key,value) ->
		(*First check if the name exist*)
		if gvar_exist key env then
			(*If it exists then retrieve the object then *)
			let gvdeclobj=NameMap.find key env in
				(*copy the type of the value *)
				match value with IntLit(_) | StrLit(_) | BoolLit(_) | AVPList(_) as literal ->
				(*Now create a new VDECL object based on the new value for an existing global variable*)
					let gvdeclupdobj = {vtype=gvdeclobj.vtype;vname=gvdeclobj.vname;value=literal} in 
						(*check if the new object is consistent i.e. the value is of the compatible data type*)
						if check_gvdecl_consistency  gvdeclupdobj then
							(*if compatability check passes update and return the updated symbol table*)
							let updenv=NameMap.add key gvdeclupdobj env in updenv
							(*else clause is just to complete the if/else statement however the execution will never come to this point
							since if there is no compatability check_gvdecl_consistency will raise an exception*)
						else
							raise(Failure("Incompatible value for " ^ gvdeclobj.vname))
				|
				Id(identifier) -> 
					(*check now if identifier exist *)
					if gvar_exist identifier env then
						(*check if the identifier exist*)
						let  gvdeclobj2=NameMap.find identifier env in
							(*create a gvdecl object with the name of existing identifer and the value given in the assignment expression*)
							let gvdeclupdobj={vtype=gvdeclobj.vtype;vname=gvdeclobj.vname;value=gvdeclobj2.value} in
								(*check if the new value will be consistent with the datatype of the existing identifier*)
								if check_gvdecl_consistency  gvdeclupdobj then
									(*if yes then update the environment*)
									let updenv=NameMap.add key gvdeclupdobj env in updenv
								else
									(*raise exception if the value is inconsistent*)
									raise(Failure("Incompatible value for " ^ gvdeclobj.vname))

					else	(*raise exception if given identifer does not exist*)
						raise(Failure("Variable " ^ identifier ^ "does not exist"))

				|			
				(*Nested assignment not allowed*)
				LitAssign(key2,value2)  ->
					raise(Failure("Invalid assignment " ^key ^"=" ^key2))

		else	(*If the variable does not exist raise an exception*)
			raise(Failure("Variable " ^ key ^ "does not exist"))

	|

	_  -> raise(Failure("Invalid assignment type"))


(************************************************************************
Name = check_insert_colnames

Description= This method is used to check if columnnames for insert
statement are valid

Input = A list of columnnames

Output = True if columnnames are valid else exceptions are raised
*************************************************************************)
let check_insert_colnames col_names=
	if List.mem "name" col_names=true then
		if List.mem "node" col_names=true then
			if List.mem "ingressPort" col_names=true then
				if List.mem "priority" col_names=true then
					if List.mem "etherType" col_names=true then
						if List.mem "actions" col_names=true then
							true
						else
							raise(Failure("Mandatory column action missing"))
					else
						raise(Failure("Mandatory column ether type missing"))
				else
					raise(Failure("Mandatory column priority missing"))
			else
				raise(Failure("Mandatory column ingress port missing"))
		else
			raise(Failure("Mandatory column node missing"))
	else
		raise(Failure("Mandatory column name missing"))

(************************************************************************
Name = find_dup

Description= This method is used to check if there are any duplicate
columns

Input = A list of columnnames

Output = False if columnnames are non duplicate else exception is raised
*************************************************************************)

let rec find_dup= function 
[] -> 
	false
|
hd::tl -> 
	if List.mem hd tl then
		raise(Failure("Column " ^hd ^ " appears multiple times, duplicates are not allowed in NwQL"))
	else
		find_dup tl


(************************************************************************
Name = find_dup_expr

Description= This method is used to check if there are any duplicate
expressions

Input = A list of expressions

Output = False if expressions are non duplicate else exception is raised
*************************************************************************)
let rec find_dup_expr= function
[] ->
        false
|
hd::tl ->
        if List.mem hd tl then
                raise(Failure("Flow name " ^(ptyexpr hd) ^ " appears multiple times, duplicate flownames are not allowed in DELETE statements"))
        else
                find_dup_expr tl
(************************************************************************
Name = check_colnames

Description= This method is used to check if columnnames are valid

Input = A list of columnnames 

Output = True if columnnames are valid else exceptions are raised
*************************************************************************)

let check_colnames col_names =

	let val_col_names =
		[
		"actions";"etherType";"hardTimeout"; 
		"idleTimeout";"ingressPort";"name";"node";
		"nwDst";"nwSrc";"priority";"protocol";
		"tpDst";"tpSrc";"vlanId";"vlanPriority"
		] 
		in
			let rec cmp_colNames list_of_cnames=
				match list_of_cnames with
				(*This is only possible when the
				entire list is traversed of an empty
				list is given as an input*)
				[]->true
				|
				hd::tl ->
					if List.mem hd val_col_names then
						cmp_colNames tl
					else
						raise(Failure("Invalid Column name " ^ hd))

			in cmp_colNames col_names
(************************************************************************
Name = check_tablename

Description= This method is used to check if provided tablename is valid

Input = Tablename

Output = True if tablename is valid else False
*************************************************************************)
let check_tablename tablename =
	let tableRegex=Regex.create_exn "\\'http:\\/\\/\\b([0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3})\\b:\\b(655[0-3][0-5]|655[0-2][0-9]|6[0-5][0-4][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[0-9])\\b\\/controller\\/nb\\/v2\\/flowprogrammer\\/default\\',auth=\\('[_a-zA-Z]+','[_a-zA-Z]+'\\)" in
	Regex.matches tableRegex tablename
(************************************************************************
Name = check_identifier

Description= This method is used to check values of columns are
of compatible expression types when values are global variables

Input = variablename,environment,datatype and columnname

Output = True if is consistent else the mutually recursive method will
cause an exception to happen
*************************************************************************)
let rec check_identifier vname env datatype cname =
	if gvar_exist vname env then
		let literal=gvar_value vname env in
			match literal with datatype -> check_value (env,cname,literal.value)
	else  raise(Failure("Variable " ^vname^ " does not exist"))

and 	

(************************************************************************
Name = check_value

Description= This method is used to check values of columns are
of compatible expression types when values are literals of expression type

Input = environment,columnname and expression

Output = True if is consistent else this method will
cause an exception to happen
*************************************************************************)

check_value (env,colname,expr) =
	(*some regular expressions used throughout*)
        let ipaddrregex =Regex.create_exn "\"\\b(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\b\""
	and
        portaddrregex=Regex.create_exn "\\b(655[0-3][0-5]|655[0-2][0-9]|6[0-5][0-4][0-9][0-9]|[1-5][0-9][0-9][0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0            -9]|[0-9])\\b"
        and
        vlanidregex= Regex.create_exn "\\b(40[0-8][0-9]|409[0-5]|3[0-9][0-9][0-9]|2[0-9][0-9][0-9]|1[0-9][0-9][0-9]|[1-9][0-9][0-9]|[1-9][0-9]|[0-9])\\b"
        and
        vlanprioregex=Regex.create_exn "0|1|2|3|4|5|6|7"
        and
        identifierregex=Regex.create_exn "[a-zA-Z][a-zA-Z0-9_]+"
        and
        nodeidregex=Regex.create_exn ".+"
        in
	(*matching of different column names start*)
        match colname with "actions" ->
                (*a separate method for performing semantic analysis of column name actions*)
                let check_actions_value envr value =
                        (*Formally actions has literals of expression type AVP*)
                        match value with AVPList(avp) ->
                                (*A subset of AVP literals may qualify for actions column*)
                                let avpmatch=function [("\"DROP\"","\"DROP\"")] -> true
                                        |
                                        [("\"SET_VLAN_ID\"",vlanid)] ->
                                                if Regex.matches vlanidregex vlanid = true then true
                                                else raise(Failure("Bad value for SET_VLAN_ID"))
                                        |
                                        [("\"SET_VLAN_PCP\"",priority)] ->
                                                if Regex.matches vlanprioregex priority=true then true
                                                else raise (Failure("Bad value for SET_VLAN_PCP"))
                                        |
                                        [("\"SET_NW_DST\"",ipdst)] ->
                                                if Regex.matches ipaddrregex ipdst=true then true
                                                else raise (Failure("Bad value for SET_NW_DST"))
                                        |
                                        [("\"SET_NW_SRC\"",ipsrc)] ->
                                                if Regex.matches ipaddrregex ipsrc=true then true
						else raise (Failure("Bad value for SET_NW_SRC"))
                                        |
                                        [("\"SET_TP_SRC\"",portsrc)] ->
                                                if Regex.matches portaddrregex portsrc=true then true
                                                else raise (Failure("Bad value for SET_TP_SRC"))
                                        |
                                        [("\"SET_TP_DST\"",portdst)] ->
                                                if Regex.matches portaddrregex portdst=true then true
                                                else raise (Failure("Bad value for SET_TP_DST"))
                                        |
                                        (*Everything else is an exception*)
                                        _  -> raise(Failure("Value of actions is out of range")) in avpmatch avp
			|
                        (*besides AVPList the value may also be an identifier representing datatype of AVP*)
                        Id(identifier) -> 
				check_identifier identifier env "AVP" "actions" 
			|
			_ -> raise(Failure("Illegal literal type for column actions"))
		in check_actions_value env expr
	|

	"name" ->
		let check_name_value envr value=
			match value with StrLit(lit) ->
                        	if Regex.matches identifierregex lit=true then true
                                else raise(Failure("Bad value for name"))
			|
			Id(identifier) ->
				check_identifier identifier env "String" "name"
			|
			_ -> raise(Failure("Illegal literal type for column name"))
		in check_name_value env expr
	|
	"node" ->
                let check_node_value envr value=
                        match value with StrLit(lit) ->
                        if Regex.matches nodeidregex lit=true then true
                        else raise(Failure("Bad value for nodeid"))
			|
			Id(identifier) ->
				check_identifier identifier env "String" "node"
			|
			_ -> raise(Failure("Illegal literal type for column node"))
		in check_node_value env expr
	|
	"ingressPort" ->
                let check_ingressp_value envr value=
                        (*Ingress port could theoritically be any number*)
                        match value with IntLit(lit) -> true
                        |
                        Id(identifier) ->
				check_identifier identifier env "Int" "ingressPort"
			|
			_ -> raise(Failure("Illegal literal type for column ingressPort"))
		in check_ingressp_value env expr
	|
	"vlanId" ->
        	let check_vlanid_value envr value=
                        match value with IntLit(lit) ->
                                if (lit >=0 && lit<4096)
                                        then true
                                        else raise(Failure("Bad value for vlanId"))
			|
                        Id(identifier) ->
				check_identifier identifier env "Int" "vlanId"
			|
			_ -> raise(Failure("Illegal literal type for column VlanId"))
		in check_vlanid_value env expr
	|
	"priority" ->
	 	let check_flow_prio envr value=
                        match value with IntLit(lit) -> true
                        |
                        Id(identifier) ->
				check_identifier identifier env "Int" "priority"
			|
			_ -> raise(Failure("Illegal literal type for column priority"))
		in check_flow_prio env expr	
	|
	"idleTimeout" ->
        	let check_idle_to envr value=
                        match value with IntLit(lit) -> true
			|
                        Id(identifier) ->
				check_identifier identifier env "Int" "idleTimeout"
			|
			_ -> raise(Failure("Illegal literal type for column idle time out"))
		in check_idle_to env expr
	|
	"hardTimeout" ->
        	let check_hard_to envr value=
                        match value with IntLit(lit) -> true
			|
                        Id(identifier) ->
				check_identifier identifier env "Int" "hardTimeout"
			|
			_ -> raise(Failure("Illegal literal type for column hard time out"))
		in check_hard_to env expr
	|
	"vlanPriority" ->
        	let check_vlanprio_value envr value=
                	match value with IntLit(lit) ->
                	if (lit >=0 && lit<8)
                        	then true
                	else
                        	raise(Failure("Bad value for vlan Priority"))
                	|
                        Id(identifier) ->
				check_identifier identifier env "Int" "vlanPriority"
			|
			_ -> raise(Failure("Illegal literal type for Vlan priority"))
		in check_vlanprio_value env expr
	|
	"etherType" ->
        	let check_ethertype_value envr value=
                	match value with StrLit(lit) ->
                        if (lit="\"0x800\"" || lit= "\"0x8100\"")
				then true
                        else
                                 raise(Failure("Bad value for vlan ether type"))
			|
                        Id(identifier) ->
				check_identifier identifier env "String" "etherType"
			|
			_ -> raise(Failure("Illegal literal type for etherType"))
		in check_ethertype_value env expr
	|"protocol" ->
        	let check_prototype_value envr value=
                	match value with StrLit(lit) ->
                        if (lit="\"tcp\"" || lit= "\"udp\"")
                                then true
                        else
                                 raise(Failure("Bad value for protocol type"))
			|
                        Id(identifier) ->
				check_identifier identifier env "String" "protocol"
			|
			_ -> raise(Failure("Illegal literal type for protocol"))
		in check_prototype_value env expr
	|"tpSrc" ->
        	let check_tpsrc_value envr value=
                 	match value with IntLit(lit) ->
                        if (lit>=0 && lit< 65536)
                                then true
                        else
                                raise(Failure("Bad value for Source transport port"))
                	|
                        Id(identifier) ->
				check_identifier identifier env "Int" "tpSrc"
			|
			_ -> raise(Failure("Illegal literal type for tpSrc"))
		in check_tpsrc_value env expr
	|"tpDst" ->
        	let check_tpdst_value envr value=
                 	match value with IntLit(lit) ->
                        if (lit>=0 && lit< 65536)
                                then true
                        else
                                raise(Failure("Bad value for Destination transport port"))
                	|

			Id(identifier) ->
				check_identifier identifier env "Int" "tpDst"
			|
			_ -> raise(Failure("Illegal literal type for tpDst"))
		in check_tpdst_value env expr
	|"nwSrc" ->
        	let check_nwsrc_value env value=
                	match value with StrLit(lit) ->
                        if Regex.matches ipaddrregex lit= true then true
                        else
      	                	raise(Failure("Bad value for Source IP address"))
			|
			Id(identifier) ->
				check_identifier identifier env "String" "nwSrc"
			|
			_ -> raise(Failure("Illegal literal type for nwSrc"))
		in check_nwsrc_value env expr
	|"nwDst" ->
        	let check_nwdst_value env value=
                	match value with StrLit(lit) ->
                        if Regex.matches ipaddrregex lit= true then true
                        else
      	                	raise(Failure("Bad value for Destination transport IP address"))
			|
			Id(identifier) ->
				check_identifier identifier env "String" "nwDst"
			|
			_ -> raise(Failure("Illegal literal type for nwDst"))
		in check_nwdst_value env expr
	|_ as invalid ->
		raise(Failure("Invalid column name "^invalid))
